/*
        Copyright (c) 2012, MICEX-RTS. All rights reserved.

        Plaza-2 Client Gate API Usage Sample.
        Messages sender.

        All the software and documentation included in this and any
        other MICEX-RTS CGate Releasese is copyrighted by MICEX-RTS.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted only by the terms of a valid
        software license agreement with MICEX-RTS.

        THIS SOFTWARE IS PROVIDED "AS IS" AND MICEX-RTS DISCLAIMS ALL WARRANTIES
        EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
        NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR
        PURPOSE.  MICEX-RTS DOES NOT WARRANT THAT USE OF THE SOFTWARE WILL BE
        UNINTERRUPTED OR ERROR-FREE.  MICEX-RTS SHALL NOT, UNDER ANY CIRCUMSTANCES, BE
        LIABLE TO LICENSEE FOR LOST PROFITS, CONSEQUENTIAL, INCIDENTAL, SPECIAL OR
        INDIRECT DAMAGES ARISING OUT OF OR RELATED TO THIS AGREEMENT OR THE
        TRANSACTIONS CONTEMPLATED HEREUNDER, EVEN IF MICEX-RTS HAS BEEN APPRISED OF
        THE LIKELIHOOD OF SUCH DAMAGES.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#endif//_WIN32
#include <assert.h>

#define CG_SHORT_NAMES
#include <cgate.h>


#include "forts_messages.h"

// helper macro to call CGate function
// and exit if there was error

#define CHECK_FAIL(f, msg) \
	{ \
		uint32_t res = f; \
		if (res != CG_ERR_OK) \
		{ \
			printf("Client gate error (%s): %x\n", msg, res); \
			exit(1); \
		} \
	} 

// helper macro to call CGate function
// and print warning if there was error

#define WARN_FAIL(f, msg) \
	{ \
		uint32_t res = f; \
		if (res != CG_ERR_OK) \
			printf("Error (%s): %x\n", msg, res); \
	} 

// helper variable used as a notification flag to exit
static volatile int bExit = 0;

#ifdef _WIN32
BOOL WINAPI InterruptHandler(DWORD reason)
{
	printf("\n----BREAK----\n");
	bExit = 1;
	return 1;
}
#else
#include <unistd.h>
void Sleep(int msec)
{
	usleep(msec * 1000);
}
#endif//_WIN32


const char* cfg_cli_conn = {"p2tcp://127.0.0.1:4001;app_name=test_send"};

const char* cfg_cli_publisher = "p2mq://FORTS_SRV;category=FORTS_MSG;name=srvlink;timeout=5000;scheme=|FILE|forts_messages.ini|message";

const char* cfg_cli_listener = "p2mqreply://;ref=srvlink";

cg_conn_t* cli_conn = 0;
cg_listener_t *cli_listener = 0;
cg_publisher_t *cli_publisher = 0;

CG_RESULT ClientMessageCallback(cg_conn_t* conn, cg_listener_t* listener, struct cg_msg_t* msg, void* data)
{
	switch (msg->type)
	{
	case CG_MSG_DATA: 
		{
		uint32_t* data = msg->data;
		printf("Client received reply [id:%d, data: %d, user-id: %d, name: %s]\n", 
			((struct cg_msg_data_t*)msg)->msg_id, 
			*((uint32_t*)msg->data), 
			((struct cg_msg_data_t*)msg)->user_id,
			((struct cg_msg_data_t*)msg)->msg_name);

		{
			struct scheme_desc_t* scheme;
			size_t bufSize;
			
			if (cg_lsn_getscheme(listener, &scheme) != CG_ERR_OK)
				scheme = 0;
				
			if (cg_msg_dump(msg, scheme, 0, &bufSize) == CG_ERR_BUFFERTOOSMALL)
			{					
				char* buffer = malloc(bufSize+1);
				
				bufSize++;		
				if (cg_msg_dump(msg, scheme, buffer, &bufSize) == CG_ERR_OK)
					printf("client dump: %s\n", buffer);
				free(buffer);
				
			}
		}
		break;
		}
	case CG_MSG_P2MQ_TIMEOUT: 
		{
		printf("Client reply TIMEOUT\n");
		break;
		}
	default:
		printf("Message 0x%X\n", msg->type);
	}
	return CG_ERR_OK;
}

void send_request(cg_publisher_t* pub, uint32_t counter)
{
	struct FutAddOrder* ord ;
	struct cg_msg_data_t* msg;

	CHECK_FAIL(pub_msgnew(pub, CG_KEY_NAME, "FutAddOrder", (struct cg_msg_t**)&msg), "msgnew");
	memcpy(msg->data, &counter, sizeof(counter));
	msg->user_id = counter;
	
	ord = (struct FutAddOrder*)msg->data;
	strcpy(ord->broker_code, "HB00");
	strcpy(ord->isin, "RTS-6.12");
	strcpy(ord->client_code, "000");
	ord->type = 1;
	ord->dir = 1;
	ord->amount = 1;
	strcpy(ord->price, "1700");
	ord->ext_id = 0;
		
	WARN_FAIL(pub_post(pub, (struct cg_msg_t *)msg, CG_PUB_NEEDREPLY), "pub_post");
	WARN_FAIL(pub_msgfree(pub, (struct cg_msg_t *)msg), "pub_msgfree");

}

int main()
{	
	int counter = 0;
#ifdef _WIN32
	SetConsoleCtrlHandler(InterruptHandler, 1);
#endif//_WIN32
		
	CHECK_FAIL(env_open("ini=send.ini;key=11111111"), "env_open");

	// init

	CHECK_FAIL(conn_new(cfg_cli_conn, &cli_conn), "conn_new");
	CHECK_FAIL(conn_open(cli_conn, 0), "conn_open");

	// wait for connection to become ready

	while (1)
	{
		uint32_t stateCli;
		CHECK_FAIL(conn_getstate(cli_conn, &stateCli), "conn_getstate");

		if (stateCli == CG_STATE_ERROR)
		{
			printf("Failed to open connections\n");
			exit(3);
		}

		if (stateCli == CG_STATE_ACTIVE)
			break;
	}

	CHECK_FAIL(pub_new(cli_conn, cfg_cli_publisher, &cli_publisher), "pub_new");
	CHECK_FAIL(lsn_new(cli_conn, cfg_cli_listener, &ClientMessageCallback, 0, &cli_listener), "lsn_new");

	CHECK_FAIL(pub_open(cli_publisher, 0), "pub_open");
	CHECK_FAIL(lsn_open(cli_listener, 0), "lsn_open");

	while (!bExit)
	{	
		uint32_t err;
		err = conn_process(cli_conn, 0, 0);
		if (err != CG_ERR_TIMEOUT && err != CG_ERR_OK)
			printf("Failed to process  connection: 0x%X\n", err);

		printf("Sending message to server\n");
		send_request(cli_publisher, counter++);

		Sleep(1000);
	}

	// done
	WARN_FAIL(conn_close(cli_conn), "conn_close");
	WARN_FAIL(lsn_destroy(cli_listener), "lsn_destroy");
	WARN_FAIL(pub_destroy(cli_publisher), "pub_destroy");
	WARN_FAIL(conn_destroy(cli_conn), "conn_destroy");
	
	CHECK_FAIL(env_close(), "env_close");
	return 0;
}

