{$ifndef Unix}
  {$APPTYPE CONSOLE}
{$endif}
program send;

uses {$ifdef FPC}
       {$ifdef MSWINDOWS}windows,{$endif}
       {$ifdef Unix}unix,{$endif}
     {$else}
       windows,
     {$endif}
     sysutils,
     cgate;

type  pFutAddOrder       = ^tFutAddOrder;
      tFutAddOrder       = record
        broker_code      : array[0..4] of ansichar;
        isin             : array[0..25] of ansichar;
        client_code      : array[0..3] of ansichar;
        _type            : longint;
        dir              : longint;
        amount           : longint;
        price            : array[0..17] of ansichar;
        comment          : array[0..20] of ansichar;
        broker_to        : array[0..20] of ansichar;
        ext_id           : longint;
        du               : longint;
        date_exp         : array[0..8] of ansichar;
        hedge            : longint;
        dont_check_money : longint;
        local_stamp      : tcg_time;
      end;

var   hConn, hPublisher, hReplyListner : THandle;
      status, process_res              : longint;
      fexit                            : boolean = false;
      last_send_time                   : cardinal = 0;
      counter                          : longint = 0;

function CHECK_FAIL(err_code: longint; const msg: string): boolean;
begin
  result:= (err_code = CG_ERR_OK);
  if not result then begin
    cg_log_infostr(pAnsiChar(format('Client gate error (%s): %.8x', [msg, err_code])));
    halt;
  end;
end;

function WARN_FAIL(err_code: longint; const msg: string): boolean;
begin
  result:= (err_code = CG_ERR_OK);
  if not result then cg_log_infostr(pAnsiChar(format('Client gate error (%s): %.8x', [msg, err_code])));
end;

{$ifdef MSWINDOWS}
function CtrlHandler(CtrlType: Longint): bool; stdcall;
begin
  fexit:= true;
  result:= true;
end;
{$endif}

{$ifdef Unix}
function GetTickCount: Cardinal;
var t : timeval;
begin
  fpgettimeofday(@t,nil);
  result := ((Int64(t.tv_sec) * 1000000) + t.tv_usec) div 1000;
end;
{$endif}

function ANSIToOEM(const ANSIStr: ansistring): ansistring;
var i : integer;
begin
  result:=ANSIStr;
  for i:= 1 to length(result) do
    case result[i] of
      #192..#239 : dec(result[i], 64);
      #240..#255 : dec(result[i], 16);
    end;
end;

function MessageCallback(hconn: THandle; hlistener: THandle; msg: pcg_msg; data: pointer): longint; cdecl;
var buf    : pAnsiChar;
    sz     : size_t;
    scheme : pcg_scheme_desc;
begin
  case msg^.msg_type of
    CG_MSG_DATA         : with pcg_msg_data(msg)^ do begin
                            writeln(format('Client received reply [id:%d, data: 0x%p, user-id: %d, name: %s]', [msg_id, data, user_id, string(msg_name)]));

                            if (cg_lsn_getscheme(hlistener, scheme) <> CG_ERR_OK) then scheme:= nil;

                            sz:= 0; buf:= nil;
                            if (cg_msg_dump(msg, scheme, buf, sz) = CG_ERR_BUFFERTOOSMALL) then begin
                              buf:= allocmem(sz + 1);
                              if (cg_msg_dump(msg, scheme, buf, sz) = CG_ERR_OK) then writeln('client dump: ', ANSIToOEM(string(buf)));
                              freemem(buf);
                            end;

                          end;
    CG_MSG_P2MQ_TIMEOUT : writeln('Client reply TIMEOUT');

   else                   writeln(format('Message 0x%.8x', [msg^.msg_type]));
  end;

  result:= CG_ERR_OK;
end;

procedure send_request(pub: THandle; counter: longint);
var   msg     : pcg_msg;
begin
  CHECK_FAIL(cg_pub_msgnew(pub, CG_KEY_NAME, pAnsiChar('FutAddOrder'), msg), 'cg_pub_msgnew');

  if assigned(msg) and (msg^.msg_type = CG_MSG_DATA) then try
    pcg_msg_data(msg)^.user_id:= counter;
    with pFutAddOrder(msg^.data)^ do begin
      StrPLCopy(isin, 'RTS-6.12', sizeof(isin) - 1);
      StrPLCopy(client_code, '000', sizeof(client_code) - 1);
      _type  := 1;
      dir    := 1;
      amount := 1;
      StrPLCopy(price, '155050', sizeof(price) - 1);
      ext_id:= 0;
    end;
  finally WARN_FAIL(cg_pub_post(pub, msg, CG_PUB_NEEDREPLY), 'pub_post'); end;
  
  WARN_FAIL(cg_pub_msgfree(pub, msg), 'pub_msgfree');
end;

begin
  hConn:= 0; hPublisher:= 0; hReplyListner:= 0;

  {$ifdef MSWINDOWS}
  SetConsoleCtrlHandler(@CtrlHandler, true);
  {$endif}

  try

// p2tcp://127.0.0.1:4001;app_name=test1
// p2repl://FORTS_FUTINFO_REPL
// p2repl://FORTS_FUTTRADE_REPL

    CHECK_FAIL(cg_env_open('ini=send.ini;subsystems=mq,replclient;key=11111111'), 'cg_env_open');
    while not fexit do begin
      if (hConn = 0) then begin
        WARN_FAIL(cg_conn_new('p2tcp://127.0.0.1:4001;app_name=test1', hConn), 'cg_conn_new');
      end else begin
        WARN_FAIL(cg_conn_getstate(hConn, status), 'cg_conn_getstate');
        case status of
          CG_STATE_ERROR  : WARN_FAIL(cg_conn_close(hConn), 'cg_conn_close');
          CG_STATE_CLOSED : WARN_FAIL(cg_conn_open(hConn, nil), 'cg_conn_open');
          CG_STATE_OPENING: process_res:= cg_conn_process(hConn, 1, nil);
          CG_STATE_ACTIVE : begin
                              process_res:= cg_conn_process(hConn, 1, nil);
                              if (process_res <> CG_ERR_OK) and (process_res <> CG_ERR_TIMEOUT) then
                                cg_log_infostr(pAnsiChar(format('Warning: connection state request: %d', [process_res])));

                              if (hReplyListner = 0) then begin
                                WARN_FAIL(cg_lsn_new(hConn, 'p2mqreply://;ref=srvlink', @MessageCallback, nil, hReplyListner), 'cg_lsn_new');
                              end else begin
                                WARN_FAIL(cg_lsn_getstate(hReplyListner, status), 'cg_lsn_getstate');
                                case status of
                                  CG_STATE_CLOSED : WARN_FAIL(cg_lsn_open(hReplyListner, nil), 'cg_lsn_open');
                                  CG_STATE_ERROR  : WARN_FAIL(cg_lsn_close(hReplyListner), 'cg_lsn_close');
                                end;
                              end;

                              if (hPublisher = 0) then begin
                                WARN_FAIL(cg_pub_new(hConn, 'p2mq://FORTS_SRV;category=FORTS_MSG;name=srvlink;timeout=5000;scheme=|FILE|forts_messages.ini|message', hPublisher), 'cg_pub_new');
                              end else begin
                                WARN_FAIL(cg_pub_getstate(hPublisher, status), 'cg_pub_getstate');
                                case status of
                                  CG_STATE_CLOSED : WARN_FAIL(cg_pub_open(hPublisher, nil), 'cg_pub_open');
                                  CG_STATE_ERROR  : WARN_FAIL(cg_pub_close(hPublisher), 'cg_pub_close');
                                  CG_STATE_ACTIVE : if (hReplyListner <> 0) then begin
                                                      cg_lsn_getstate(hReplyListner, status);
                                                      if (status = CG_STATE_ACTIVE) then begin
                                                        if (abs(GetTickCount - last_send_time) > 1000) then begin
                                                          send_request(hPublisher, counter);
                                                          last_send_time:= GetTickCount;
                                                          inc(counter);
                                                        end;
                                                      end;
                                                    end;
                                end;
                              end;
                            end;
        end;
      end;
    end;

    if (hReplyListner <> 0) then begin
      cg_lsn_close(hReplyListner);
      cg_lsn_destroy(hReplyListner);
    end;

    if (hPublisher <> 0) then begin
      cg_pub_close(hPublisher);
      cg_pub_destroy(hPublisher);
    end;

    if (hConn <> 0) then begin
      cg_conn_close(hConn);
      cg_conn_destroy(hConn);
    end;

    CHECK_FAIL(cg_env_close, 'cg_env_close');

  except on e: exception do begin writeln(#$0d#$0a'Exception: ', e.message); cg_log_infostr(pAnsiChar(e.message)); end; end;

  writeln(#$0d#$0a'PRESS ENTER TO EXIT'); readln;
end.
